#!/bin/bash
##################################################
## Job script for bwUniCluster                  ##
## to run M independent sub jobs on N cores     ##
##################################################
## ver.  : 2015-02-25, Robert Barthel, KIT, SCC
##
## Syntax:
##   ^                         = begin of line
##   <int>                     = integer placeholder
##   ^##                       = comment
##   
##   ^#MSUB -l nodes=<int>:ppn=<int> = defines nodes and cores per node
##   ^#MSUB -l walltime=hh:mm:ss     = defines walltime in hh:mm:ss
##   ^#MSUB -N <jobname>             = defines jobname
##   ^#MSUB -m <email_option>        = defines nofication condition
##
## Note  : 
##   - all '#MSUB -?' are equivalent to 'msub -?'
##   - for more options, see 'man msub'
##   - '# MSUB' is a comment, while '#MSUB' is a msub option statement

##################################################
## HEADER: Define here your Compute resources   ##
##################################################
#MSUB -N Pseudoparallel
#MSUB -l advres=workshop.54,nodes=2:ppn=2
#MSUB -l walltime=00:10:00
#MSUB -m n

##################################################
## FUNCTIONS:                                   ##
##################################################
usage(){
local SH=$(basename $0)
echo -e "
  \033[44;37;4;1m  H E L P  \033[0m

  USAGE     : $SH [${e1}-options${e0} <argument>]  jobfile1 jobfileN

  ${e1}-h${e0}          Print this message

  jobfile(s):  
  -----------
  * 1 non-commented line = 1 job
  * Multiple commands per line must be separated by semicolon
  * Stdout and stderr redirections must be removed, since handled by $(basename ${0})
"
}


##################################################
## INPUT:                                       ##
##################################################
## Reading options of script
while getopts ":h" OPT  ; do
   case "$OPT" in
       h) usage ; exit 0 ;;
       #-----
      \?) echo -e "\n  ERROR: option '-${OPTARG}' is unknown"; usage; exit 1 ;;
    esac
done
##
## Reading names of jobfiles and check existence
files=""; WDIR=${PWD}
i=$(($OPTIND))
while [ $i -le $# ] ; do
   eval tmp=\${$i}
   if [ -e ${WDIR}/${tmp} ] ; then
      files="${files}${tmp} "
   fi
   let i+=1
done
if [ -z "$files" ] ; then
   echo -e "\n  ERROR: no valid jobfiles given"; usage; exit 1
fi
##
## Generate ARRAY of job commands
for file in $files ; do
   ## For ARRAY go though files per line since spaces in one line are not protected
   while read -r line ; do
      EXE+=("${line}")
   done < "$file"
done


##################################################
## PROCESSING                                   ##
##################################################
echo "***********************************************"
echo "*  Pseudo parallelisation of sequential jobs  *"
echo "*            for bwUniCluster                 *"
echo "*                                             *"
echo "*     2015-02-25, Robert Barthel, KIT, SCC    *"
echo "***********************************************"
echo -e "-----\nStarting  time: $(date)" 

## 1.) For running sequential independent jobs parallel we check:
##
## 1.a) How much cores we got (-> this has to be adjusted by nodes=N:ppn=NoCore
##    -> this information comes with MOAB
NoCore=${MOAB_PROCCOUNT:-99}     ## Number of Core
dim_NoCore=$(echo "${NoCore}" |  awk '{if ($1==1000) a=4; else a=int(log($1)/log(10))+1; print a}') ## int() of awk is buggy
echo "Number of assigned Cores is: $NoCore"
##
## 1.b) Generate a slot list
i=1
slot_list=""
for host in $(srun hostname) ; do
   no=$(printf '%0'${dim_NoCore}'d' ${i})
   slot_list="${slot_list}${host}:${no} "
   let i+=1    
done
echo "Generated Slot list        : $slot_list"

## C) count the number of jobs from the number of elements in $EXE array
NoJob=${#EXE[@]}
dim_NoJob=$(echo "${NoJob}" |  awk '{if ($1==1000) a=4; else a=int(log($1)/log(10))+1; print a}') ## int() of awk is buggy
echo "Number of Jobs in Script is: $NoJob"
echo "-----"

## Before starting loop, set job counter
i=1

## Run loop as long as loop counter is smaller than maximum no. of jobs
## Note: Index of bash array starts at 0
while [ $i -le $NoJob ] ; do
   no=$(printf '%0'${dim_NoJob}'d' ${i})
   echo "Do Job no: ${no}"
   ##
   ## Build subjob script, because at the end of job a cleanup of lock file is required
   ## job slave PART A:
   job_slave=job_slave_${MOAB_JOBID:-$$}_${no}.sh
   echo "|-- pack job command and removal of LOCK file in job slave script: $job_slave"
   ##
   echo "#!/bin/bash"                           > $job_slave
   echo "## This is job slave   : ${no}"       >> $job_slave
   echo "## This is EXE element : ${EXE[i-1]}" >> $job_slave
   echo 'echo -e "Starting  time: $(date)"'    >> $job_slave
   echo "${EXE[i-1]}"                          >> $job_slave
   echo "if [ \$? -eq 0 ] ; then echo Job ${no}: successfully finshed ; fi" >> $job_slave
   ##
   ## Loop over available ressource slots/hosts/CPUs -> and use available slots to do sub processes / job slaves
   for host in $slot_list ; do
      if [ ! -e ${MOAB_SUBMITDIR}/lock_${host} ] ; then
         ## Remove numbering from hostname
         host_wo_no="$(echo $host | cut -d: -f1)"
         ##
         ## Lock slot
         lock_file=${MOAB_SUBMITDIR}/lock_${host}
         touch ${lock_file}
         ##
         ## Now we have the information to finish job slave PART B:
         err=${MOAB_SUBMITDIR}/${job_slave%.sh}_${host}.err
         log=${err%err}log
         echo "if [ ! -s ${err} ] ; then rm -f ${err} ; fi" >> $job_slave
         echo "echo Job ${no}: remove ${lock_file}"         >> $job_slave
         echo "rm -f ${lock_file}"                          >> $job_slave
         echo 'echo -e "Finishing time: $(date)"'           >> $job_slave
         ##
         ## Make job slave executable
         chmod 700 $job_slave
         ## 
         ## Print pseudo_parallel_cmd before executing
         ## -> current assignment of one job per core is done by SLURM and its srun command
         pseudo_parallel_cmd="srun -N 1 -n 1 -w ${host_wo_no} ${MOAB_SUBMITDIR}/${job_slave}"
         echo "|-- execute command: ${pseudo_parallel_cmd} > ${log} 2> ${err}"
         echo "  |-- at           : $(date)"
         ${pseudo_parallel_cmd} > "${log}" 2> "${err}" & 
         ##
         ## increase job counter
         let i+=1
         ##
         ## prevent that this job slave is executed multiple times
         break
      fi
   done
   sleep 2
   ##
   ## Wait until a new slots is available, i.e. each job removes its LOCK file after being finished
   echo "|-- wait until next slot is free"
   while [ $(ls -1 ${MOAB_SUBMITDIR}/lock_* 2>/dev/null | wc -l) -ge ${NoCore} ] ; do
      sleep 1
   done
   ##
done

## Wait until all slots are available -> then terminate master script
while [ $(ls -1 ${MOAB_SUBMITDIR}/lock_* 2>/dev/null | wc -l) -ne 0 ] ; do
   sleep 1
done
sleep 1

echo "-----"
echo "All Jobs done"
echo "Finishing time: $(date)"
exit 0
